

class ResultAccumulator:

    def __init__(self, *, output_manager):
        self.output_manager = output_manager
        self.candidates = []

    def add_entry(self, entry):
        entry = self._select_entry(entry)
        self._output_found(entry)
        self.candidates.append(entry)

    async def revalidate(self, validator):
        for entry in self.candidates:
            if await validator.is_valid(entry):
                self._output_found(entry, confirmed=True)

    def _output_found(self, entry, **kwargs):
        if "file" in entry.arguments or "path" in entry.arguments:
            data = self._get_data(entry, kwargs)
            message = self._format_message(entry, data)
            self.output_manager.output_result(message, data=data)

    def _format_message(self, entry, data):
        url = entry.request.url
        return "{prefix}{desc} at: {url}{suffix}".format(prefix=self._get_prefix(entry, data),
                                                         suffix=self._get_suffix(entry, data),
                                                         desc=data["description"], url=url)

    def _get_prefix(self, entry, data):
        if "file" in entry.arguments:
            if entry.response.code == 500:
                return "ISE, "
            elif len(entry.response.raw) == 0:
                return "Empty "
        if "path" in entry.arguments:
            if entry.response.code == 401:
                return "Password Protected - "
            elif entry.response.code == 403:
                return "*Forbidden* "
            elif data.get("special") == "tomcat-redirect":
                return "Tomcat redirect, "
            elif entry.response.code == 500:
                return "ISE, "

        return ""

    def _get_suffix(self, entry, data):
        parts = []

        if data.get("confirmed"):
            parts.append("Confirmed")

        if data.get('har'):
            parts.append("HAR: %s" % data.get('har'))

        if parts:
            return " (%s)" % ", ".join(parts)
        else:
            return ""

    def _get_data(self, entry, additional):
        url = entry.request.url
        descriptor = entry.arguments.get("file") or entry.arguments.get("path")

        data = {"url": url,
                "description": descriptor["description"],
                "code": entry.response.code,
                "severity": descriptor.get('severity', "warning")}
        data.update(additional)

        if entry.response.code == 404 and self._detect_tomcat_fake_404(entry.response.raw):
            data["special"] = "tomcat-redirect"

        har = getattr(entry.result, "har_location", None)
        if har is not None:
            data["har"] = har

        return data

    def _detect_tomcat_fake_404(self, content):
        """ An apache setup will issue a 404 on an existing path if there is a tomcat trying to handle jsp on the same
            host """
        if content.find(b'Apache Tomcat/') != -1:
            return True

        return False

    def _select_entry(self, entry):
        if entry.result.redirects:
            last_step = entry.result.redirects[-1]
            last_step.arguments = entry.arguments
            return last_step
        else:
            return entry



# Tachyon - Fast Multi-Threaded Web Discovery Tool
# Copyright (c) 2011 Gabriel Tremblay - initnull hat gmail.com
# Copyright (C) 2018-  Delve Labs inc.
#
# GNU General Public Licence (GPL)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA

import json
import logging
from datetime import datetime

import click

from tachyon import conf
from tachyon.__version__ import __version__

FOUND = logging.INFO + 5
TIMEOUT = logging.WARNING + 5
logging.addLevelName(FOUND, "FOUND")
logging.addLevelName(TIMEOUT, "TIMEOUT")


class OutputManager:

    def output_result(self, text, data=None):
        self._add_output(text, FOUND, data)

    def output_info(self, text):
        self._add_output(text, logging.INFO)

    def output_error(self, text):
        self._add_output(text, logging.ERROR)

    def output_timeout(self, text):
        self._add_output(text, TIMEOUT)

    def output_raw_message(self, message):
        click.echo(message)

    def output_header(self):
        pass

    def flush(self):
        raise NotImplementedError

    def _add_output(self, text, level, data=None):
        raise NotImplementedError

    def _format_output(self, time, level_name, text, data):
        raise NotImplementedError

    def _get_current_time(self):
        return str(datetime.now().strftime("%H:%M:%S"))


class JSONOutput(OutputManager):

    def __init__(self):
        self.buffer = []

    def flush(self):
        self.output_raw_message(json.dumps({"result": self.buffer, "version": __version__, "from": conf.name}))

    def _add_output(self, text, level, data=None):
        formatted = self._format_output(self._get_current_time(), logging.getLevelName(level), text, data)
        self.buffer.append(formatted)

    def _format_output(self, time, level_name, text, data):
        output = {"type": level_name.lower(), "text": text, "time": time}
        if data is not None:
            output.update(**data)
        return output


class PrettyOutput(OutputManager):

    def flush(self):
        pass

    def output_header(self):
        """ Print a _cute_ program header """
        header = "\n\t Tachyon v%s - Fast Multi-Threaded Web Discovery Tool\n\t https://github.com/delvelabs/tachyon\n"
        self.output_raw_message(header % __version__)

    def _add_output(self, text, level, data=None):
        formatted = self._format_output(self._get_current_time(), logging.getLevelName(level), text, data)
        self.output_raw_message(formatted)

    def _format_output(self, time, level_name, text, data):
        output_format = "[{time}] [{level}] {text}"
        output = output_format.format(time=time, level=level_name, text=text)
        return output



# Tachyon - Fast Multi-Threaded Web Discovery Tool
# Copyright (c) 2011 Gabriel Tremblay - initnull hat gmail.com
# Copyright (C) 2018-  Delve Labs inc.
#
# GNU General Public Licence (GPL)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA
#
from tachyon import conf, textutils, database


async def execute(hammertime):
    """ This plugin process the hostname to generate host and filenames relatives to it """
    target = conf.target_host

    # Remove char to figure out the human-likely expressed domain name
    # host.host.host.com = hosthosthost.com. host.com hostcom, host, /host.ext
    # We don't test for domain.dom/domain since "cp * ./sitename" is unlikely to happen (questionable)
    added = 0

    # http://oksala.org -> oksala.org
    target = target.replace('http://', '')
    target = target.replace('https://', '')

    # Remove subpath
    first_slash = target.find('/')
    if first_slash > 0:
        target = target[0:first_slash]

    target = target.replace('/', '')
    new_target = conf.path_template.copy()
    new_target['url'] = target
    new_target['description'] = "HostProcessor generated filename"
    database.files.append(new_target)
    added += 1

    # www.oksala.org -> oksala.org
    target = target.replace('www.', '')
    new_target = conf.path_template.copy()
    new_target['url'] = target
    new_target['description'] = "HostProcessor generated filename"
    database.files.append(new_target)
    added += 1

    # oksala.org -> oksala
    dom_pos = target.rfind('.')
    nodom_target = target[0:dom_pos]
    new_target = conf.path_template.copy()
    new_target['url'] = nodom_target
    new_target['description'] = "HostProcessor generated filename"
    database.files.append(new_target)
    added += 1

    # shortdom (blabla.ok.ok.test.com -> test)
    new_target = conf.path_template.copy()
    dom_pos = target.rfind('.')
    if dom_pos > 0:
        nodom_target = target[0:dom_pos]
        start_pos = nodom_target.rfind('.')
        if start_pos > 0:
            short_dom = nodom_target[start_pos+1:]
        else:
            short_dom = nodom_target

        new_target['url'] = short_dom
        new_target['description'] = "HostProcessor generated filename"
        database.files.append(new_target)
        added += 1

        new_target = new_target.copy()
        new_target['url'] = short_dom + 'admin'
        database.files.append(new_target)
        added += 1

        new_target = new_target.copy()
        new_target['url'] = short_dom + '-admin'
        database.files.append(new_target)
        added += 1

    # flatten subdomains
    target = target.replace('.', '')
    new_target = conf.path_template.copy()
    new_target['url'] = target
    new_target['description'] = "HostProcessor generated filename"
    database.files.append(new_target)
    added += 1

    textutils.output_info(" - HostProcessor Plugin: added " + str(added) + " new filenames")



import asyncio
import json
import uuid
from http.client import responses
from os.path import join

from marshmallow_har import HAR, Entry, Request, Response, Content, Header, Creator
from tachyon.__version__ import __version__


class FileWriter:

    def __init__(self, output_dir):
        self.dir = output_dir

    def __call__(self, har: HAR):
        filename = "%s.har" % uuid.uuid4()
        file_path = join(self.dir, filename)
        with open(file_path, "w") as fp:
            fp.write(json.dumps(har.dump().data, indent=4))
        return file_path


class StoreHAR:

    def __init__(self, writer):
        self.writer = writer
        self.converter = HammerTimeToHAR()

    async def on_request_successful(self, entry):
        await asyncio.get_event_loop().run_in_executor(None, self._write_har, entry)

    def _write_har(self, entry):
        har = self.converter.convert_entries([entry])
        entry.result.har_location = self.writer(har)


class HammerTimeToHAR:

    def convert_entries(self, entries):
        out = HAR(creator=Creator(name="Tachyon", version=__version__))
        for e in entries:
            if e.result.redirects:
                for sub in e.result.redirects:
                    out.entries.append(self.convert_entry(sub))
            else:
                out.entries.append(self.convert_entry(e))
        return out

    def convert_entry(self, entry):
        out = Entry(request=self.convert_request(entry.request))

        if entry.response:
            out.response = self.convert_response(entry.response)

        return out

    def convert_headers(self, headers):
        return [Header(name=k, value=v) for k, v in headers.items()]

    def convert_request(self, request):
        out = Request(method=request.method, url=request.url,
                      headers=self.convert_headers(request.headers))

        return out

    def convert_response(self, response):
        out = Response(status=response.code, status_text=self._lookup_code_text(response.code),
                       headers=self.convert_headers(response.headers))
        out.content = Content(text=self._get_content(response))

        headers = {k.lower(): v for k, v in response.headers.items()}
        out.content.mime_type = headers.get('content-type')
        out.content.size = int(headers.get('size', -1))

        return out

    def _get_content(self, response):
        try:
            return response.content
        except UnicodeDecodeError:
            return response.raw.decode("utf-8", "surrogateescape")

    @staticmethod
    def _lookup_code_text(code):
        return responses.get(code) or "Unknown"



# Tachyon - Fast Multi-Threaded Web Discovery Tool
# Copyright (c) 2011 Gabriel Tremblay - initnull hat gmail.com
#
# GNU General Public Licence (GPL)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA
#

# Add all path plugins here
__all__ = []



import re
from difflib import SequenceMatcher
from urllib.parse import urlparse

from hammertime.rules.redirects import valid_redirects
from hammertime.ruleset import RejectRequest


class RedirectLimiter:

    def __init__(self, sequence_matching=True):
        self.sequence_matching = sequence_matching
        self.digits = re.compile(r'\d+')
        self.not_found = re.compile(r'not[-_]*found', re.I)

    async def after_headers(self, entry):
        if entry.response.code not in valid_redirects:
            return

        location = entry.response.headers.get("location")
        if not location:
            return

        if "404" in self.digits.findall(location):
            raise RejectRequest("Redirection to error page.")
        if self.not_found.search(location):
            raise RejectRequest("Redirection to error page.")

        if self.sequence_matching and self._sequences_differ(entry.request.url, location):
            raise RejectRequest("Redirection to unrelated path.")

    def _sequences_differ(self, request, redirect):
        a = urlparse(request).path
        b = urlparse(redirect).path

        matcher = SequenceMatcher(isjunk=None, a=a, b=b, autojunk=False)
        return round(matcher.ratio(), 1) <= 0.8



# Tachyon - Fast Multi-Threaded Web Discovery Tool
# Copyright (c) 2011 Gabriel Tremblay - initnull hat gmail.com
# Copyright (C) 2018-  Delve Labs inc.
#
# GNU General Public Licence (GPL)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA


from aiohttp import ClientSession, TCPConnector
from aiohttp.cookiejar import DummyCookieJar
from hammertime import HammerTime
from hammertime.config import custom_event_loop
from hammertime.engine import AioHttpEngine
from hammertime.kb import KnowledgeBase
from hammertime.ruleset import StopRequest
from hammertime.rules import DetectSoft404, RejectStatusCode, DynamicTimeout, RejectCatchAllRedirect, FollowRedirects, \
    SetHeader, DeadHostDetection, FilterRequestFromURL, DetectBehaviorChange, IgnoreLargeBody

from tachyon import conf
from tachyon.heuristics import RejectIgnoredQuery, LogBehaviorChange, MatchString, RedirectLimiter, StripTag
from tachyon.filefetcher import ValidateEntry

heuristics_with_child = []
initial_limit = 5120
default_user_agent = 'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) ' \
                     'Chrome/41.0.2228.0 Safari/537.36'


async def configure_hammertime(proxy=None, retry_count=3, cookies=None, **kwargs):
    loop = custom_event_loop()
    engine = AioHttpEngine(loop=loop, verify_ssl=False, proxy=proxy)
    await engine.session.close()
    connector = TCPConnector(loop=loop, verify_ssl=False, use_dns_cache=True, ttl_dns_cache=None)
    if cookies is not None:
        engine.session = ClientSession(loop=loop, connector=connector, cookie_jar=DummyCookieJar(loop=loop))
    else:
        engine.session = ClientSession(loop=loop, connector=connector)
    kb = KnowledgeBase()
    hammertime = HammerTime(loop=loop, request_engine=engine, retry_count=retry_count, proxy=proxy, kb=kb)
    setup_hammertime_heuristics(hammertime, **kwargs)
    hammertime.collect_successful_requests()
    hammertime.kb = kb
    return hammertime


def setup_hammertime_heuristics(hammertime, *,
                                user_agent=default_user_agent, vhost=None, confirmation_factor=1,
                                har_output_dir=None):
    global heuristics_with_child
    dead_host_detection = DeadHostDetection(threshold=200)
    detect_soft_404 = DetectSoft404(distance_threshold=6, confirmation_factor=confirmation_factor)
    follow_redirects = FollowRedirects()
    heuristics_with_child = [RejectCatchAllRedirect(), follow_redirects,
                             RejectIgnoredQuery()]
    hosts = (vhost, conf.target_host) if vhost is not None else conf.target_host

    init_heuristics = [SetHeader("User-Agent", user_agent),
                       SetHeader("Host", vhost if vhost is not None else conf.target_host),
                       dead_host_detection,
                       RejectStatusCode({503, 508}, exception_class=StopRequest),
                       StripTag('input'), StripTag('script')]

    global_heuristics = [RejectStatusCode({404, 406, 502}),
                         DynamicTimeout(1.0, 5),
                         RedirectLimiter(),
                         FilterRequestFromURL(allowed_urls=hosts),
                         IgnoreLargeBody(initial_limit=initial_limit)]

    # Dead host detection must be first to make sure there is no skipped after_headers
    hammertime.heuristics.add_multiple(init_heuristics)

    # General
    hammertime.heuristics.add_multiple(global_heuristics)
    hammertime.heuristics.add_multiple(heuristics_with_child)
    hammertime.heuristics.add_multiple([
        detect_soft_404,
        MatchString(),
        ValidateEntry(),
        DetectBehaviorChange(buffer_size=100),
        LogBehaviorChange(),
        ValidateEntry(),
    ])
    detect_soft_404.child_heuristics.add_multiple(init_heuristics)
    detect_soft_404.child_heuristics.add_multiple(heuristics_with_child)

    for heuristic in heuristics_with_child:
        heuristic.child_heuristics.add_multiple(init_heuristics)
        heuristic.child_heuristics.add_multiple(global_heuristics)

    if har_output_dir is not None:
        from tachyon.har import StoreHAR, FileWriter
        hammertime.heuristics.add(StoreHAR(writer=FileWriter(har_output_dir)))


def add_http_header(hammertime, header_name, header_value):
    set_header = SetHeader(header_name, header_value)
    hammertime.heuristics.add(set_header)
    for heuristic in heuristics_with_child:
        heuristic.child_heuristics.add(set_header)


def set_cookies(hammertime, cookies):
    add_http_header(hammertime, "Cookie", cookies)



# Tachyon - Fast Multi-Threaded Web Discovery Tool
# Copyright (c) 2011 Gabriel Tremblay - initnull hat gmail.com
#
# GNU General Public Licence (GPL)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA
#
import json
import os.path as osp
import sys

from tachyon import textutils


def _get_data_dir():
    return osp.join(osp.dirname(sys.modules['tachyon'].__file__), 'data/')


def load_json_resource(name):
    data_path = osp.join(_get_data_dir(), name + '.json')
    return load_targets(data_path)


def load_targets(file):
    """ Load the list of target paths """
    loaded = []
    with open(file) as fp:
        try:
            data = json.load(fp)
            for section in data:
                loaded.extend(section["data"])
        except json.JSONDecodeError as e:
            textutils.output_error("Error when loading file %s: %s" % (file, str(e)))
    return loaded


def load_cookie_file(filename):
    try:
        with open(filename, 'r') as cookie_file:
            content = cookie_file.read()
            content = content.replace('Cookie: ', '')
            content = content.replace('\n', '')
            return content
    except IOError:
        textutils.output_info('Supplied cookie file not found, will use server provided cookies')
        return None



# Tachyon - Fast Multi-Threaded Web Discovery Tool
# Copyright (c) 2011 Gabriel Tremblay - initnull hat gmail.com
# Copyright (C) 2018-  Delve Labs inc.
#
# GNU General Public Licence (GPL)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA


import asyncio
from urllib.parse import urlparse

import click
import tachyon.database as database
import tachyon.loaders as loaders
import tachyon.textutils as textutils
from hammertime.http import Entry
from hammertime.rules import RejectStatusCode
from hammertime.rules.deadhostdetection import OfflineHostException
from hammertime.rules.simhash import Simhash
from hammertime.ruleset import RejectRequest, StopRequest
from tachyon.directoryfetcher import DirectoryFetcher
from tachyon.filefetcher import FileFetcher

import tachyon.conf as conf
from tachyon.config import configure_hammertime, set_cookies, default_user_agent, custom_event_loop
from tachyon.generator import PathGenerator, FileGenerator
from tachyon.plugins import host, file
from tachyon.result import ResultAccumulator


def load_target_paths():
    """ Load the target paths in the database """
    textutils.output_info('Loading target paths')
    database.paths += loaders.load_json_resource('paths')


def load_target_files():
    """ Load the target files in the database """
    textutils.output_info('Loading target files')
    database.files += loaders.load_json_resource('files')


async def get_session_cookies(hammertime):
    try:
        """ Fetch the root path in a single request so aiohttp will use the returned cookies in all future requests. """
        textutils.output_info('Fetching session cookie')
        path = '/'
        await hammertime.request(conf.base_url + path)
    except RejectRequest:
        textutils.output_info('Request for website root failed.')


async def test_paths_exists(hammertime, *, recursive=False, depth_limit=2, accumulator):
    """
    Test for path existence using http codes and computed 404
    Turn off output for now, it would be irrelevant at this point.
    """

    check_closed(hammertime)

    path_generator = PathGenerator()
    fetcher = DirectoryFetcher(conf.base_url, hammertime, accumulator=accumulator)

    paths_to_fetch = path_generator.generate_paths(use_valid_paths=False)

    if len(paths_to_fetch) > 0:
        textutils.output_info('Probing %d paths' % len(paths_to_fetch))

    await fetcher.fetch_paths(paths_to_fetch)

    if recursive:
        recursion_depth = 0
        while recursion_depth < depth_limit:
            recursion_depth += 1
            paths_to_fetch = path_generator.generate_paths(use_valid_paths=True)
            await fetcher.fetch_paths(paths_to_fetch)

    count = len(database.valid_paths) - 1  # Removing one as it is the root path
    textutils.output_info('Found %d valid paths' % count)


async def load_execute_host_plugins(hammertime):
    """ Import and run host plugins """
    count = len(host.__all__)
    if count == 0:
        return

    textutils.output_info('Executing %d host plugins' % count)
    for plugin_name in host.__all__:
        plugin = __import__("tachyon.plugins.host." + plugin_name, fromlist=[plugin_name])
        if hasattr(plugin, 'execute'):
            await plugin.execute(hammertime)


def load_execute_file_plugins():
    """ Import and run path plugins """
    count = len(file.__all__)
    if count == 0:
        return

    textutils.output_info('Executing %d file plugins' % count)
    for plugin_name in file.__all__:
        plugin = __import__("tachyon.plugins.file." + plugin_name, fromlist=[plugin_name])
        if hasattr(plugin, 'execute'):
            plugin.execute()


async def test_file_exists(hammertime, accumulator, skip_root=False):
    """ Test for file existence using http codes and computed 404 """

    check_closed(hammertime)

    fetcher = FileFetcher(conf.base_url, hammertime, accumulator=accumulator)
    generator = FileGenerator()
    files_to_fetch = generator.generate_files(skip_root=skip_root)

    count = len(files_to_fetch)

    textutils.output_info('Probing %d files' % count)
    if len(database.valid_paths) > 0:
        hammertime.heuristics.add(RejectStatusCode({401, 403}))
        await fetcher.fetch_files(files_to_fetch)


def format_stats(stats):
    message = "Statistics: Requested: {}; Completed: {}; Duration: {:.0f} s; Retries: {}; Request rate: {:.2f}"
    return message.format(stats.requested, stats.completed, stats.duration, stats.retries, stats.rate)


def check_closed(hammertime):
    if hammertime.is_closed or getattr(hammertime, "_interrupted", False):
        raise KeyboardInterrupt()


async def drain(hammertime):
    iterator = hammertime.successful_requests()

    # Really make sure we are done (issue in hammertime 0.5.1 when first request is a failure?)
    while iterator.has_pending():
        async for _ in iterator:  # noqa: F841
            pass  # Just drain the pre-probe queries from the queue


async def scan(hammertime, *, accumulator,
               cookies=None, directories_only=False, files_only=False, plugins_only=False,
               **kwargs):

    if cookies is not None:
        set_cookies(hammertime, cookies)
    else:
        await get_session_cookies(hammertime)

    await load_execute_host_plugins(hammertime)

    await drain(hammertime)

    if not plugins_only:
        if not directories_only:
            textutils.output_info('Generating file targets for target root')
            load_execute_file_plugins()
            await test_file_exists(hammertime, accumulator=accumulator)

        if not files_only:
            await test_paths_exists(hammertime, accumulator=accumulator, **kwargs)

            if not directories_only:
                textutils.output_info('Generating file targets')
                load_execute_file_plugins()
                await test_file_exists(hammertime, accumulator=accumulator, skip_root=True)

    check_closed(hammertime)

    validator = ReFetch(hammertime)
    if await validator.is_valid(Entry.create(conf.base_url + "/")):
        textutils.output_info("Re-validating prior results.")
        await accumulator.revalidate(validator)
    else:
        textutils.output_error("Re-validation aborted. Target no longer appears to be up.")

    check_closed(hammertime)


class ReFetch:
    def __init__(self, hammertime):
        self.hammertime = hammertime

    async def is_valid(self, entry):
        value = getattr(entry.result, 'error_simhash', None)
        if value is not None:
            # Revalidate the responses with the known bad behavior signatures.
            current = Simhash(value)
            if any(current.distance(Simhash(k)) < 5 for k in self.hammertime.kb.bad_behavior_response):
                return False

        try:
            await self.hammertime.request(entry.request.url, arguments=entry.arguments)
            return True
        except Exception as e:
            return False


@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.option("-a", "--allow-download", is_flag=True)
@click.option("-c", "--cookie-file", default="")
@click.option("-l", "--depth-limit", default=2)
@click.option("-s", "--directories-only", is_flag=True)
@click.option("-f", "--files-only", is_flag=True)
@click.option("-j", "--json-output", is_flag=True)
@click.option("-m", "--max-retry-count", default=3)
@click.option("-z", "--plugins-only", is_flag=True)
@click.option("-x", "--plugin-settings", multiple=True)
@click.option("-p", "--proxy", default="")
@click.option("-r", "--recursive", is_flag=True)
@click.option("-u", "--user-agent", default=default_user_agent)
@click.option("-v", "--vhost", type=str, default=None)
@click.option("-C", "--confirmation-factor", default=1)
@click.option("--har-output-dir", default=None)
@click.argument("target_host")
def main(*, target_host, cookie_file, json_output, max_retry_count, plugin_settings, proxy, user_agent, vhost,
         depth_limit, directories_only, files_only, plugins_only, recursive, allow_download, confirmation_factor,
         har_output_dir):

    output_manager = textutils.init_log(json_output)
    output_manager.output_header()

    # Ensure the host is of the right format and set it in config
    parsed_url = urlparse(target_host)
    if not parsed_url.scheme:
        parsed_url = urlparse("http://%s" % target_host)

    if not parsed_url:
        output_manager.output_error("Invald URL provided.")
        return

    # Set conf values
    conf.target_host = parsed_url.netloc
    conf.base_url = "%s://%s" % (parsed_url.scheme, parsed_url.netloc)

    hammertime = None
    accumulator = ResultAccumulator(output_manager=output_manager)

    output_manager.output_info('Starting Discovery on ' + conf.base_url)

    conf.allow_download = allow_download
    for option in plugin_settings:
        plugin, value = option.split(':', 1)
        conf.plugin_settings[plugin].append(value)

    try:
        root_path = conf.path_template.copy()
        root_path['url'] = '/'
        database.valid_paths.append(root_path)
        load_target_paths()
        load_target_files()
        conf.cookies = loaders.load_cookie_file(cookie_file)
        conf.user_agent = user_agent
        conf.proxy_url = proxy
        conf.forge_vhost = vhost
        loop = custom_event_loop()
        hammertime = loop.run_until_complete(
            configure_hammertime(cookies=conf.cookies, proxy=conf.proxy_url, retry_count=max_retry_count,
                                 user_agent=conf.user_agent, vhost=conf.forge_vhost,
                                 confirmation_factor=confirmation_factor,
                                 har_output_dir=har_output_dir))
        loop.create_task(stat_on_input(hammertime))
        loop.run_until_complete(scan(hammertime, accumulator=accumulator,
                                     cookies=conf.cookies, directories_only=directories_only,
                                     files_only=files_only, plugins_only=plugins_only, depth_limit=depth_limit,
                                     recursive=recursive))

        output_manager.output_info('Scan completed')

    except (KeyboardInterrupt, asyncio.CancelledError):
        output_manager.output_error('Keyboard Interrupt Received')
    except (OfflineHostException, StopRequest):
        output_manager.output_error("Target host seems to be offline.")
    except ModuleNotFoundError as e:
        output_manager.output_error("Additional module is required for the requested options: %s" % e)
    finally:
        if hammertime is not None:
            textutils.output_info(format_stats(hammertime.stats))

        output_manager.flush()


async def stat_on_input(hammertime):
    import sys
    from datetime import datetime, timedelta

    if sys.stdin is None or not sys.stdin.readable() or not sys.stdin.isatty():
        return

    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader()
    reader_protocol = asyncio.StreamReaderProtocol(reader)

    await loop.connect_read_pipe(lambda: reader_protocol, sys.stdin)

    expiry = datetime.now()
    while True:
        await reader.readline()

        # Throttle stats printing
        if expiry < datetime.now():
            textutils.output_info(format_stats(hammertime.stats))
            expiry = datetime.now() + timedelta(seconds=2)

        if sys.stdin.seekable():
            sys.stdin.seek(-1, sys.SEEK_END)


if __name__ == "__main__":
    main()



# Tachyon - Fast Multi-Threaded Web Discovery Tool
# Copyright (c) 2011 Gabriel Tremblay - initnull hat gmail.com
# Copyright (C) 2018-  Delve Labs inc.
#
# GNU General Public Licence (GPL)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA


from uuid import uuid4
from urllib.parse import urlparse
from hammertime.http import Entry
from hammertime.ruleset import RejectRequest
from hammertime.rules.simhash import Simhash, DEFAULT_FILTER
import hashlib


class RejectIgnoredQuery:

    def __init__(self, match_treshold=5, match_filter=DEFAULT_FILTER, token_size=4):
        self.engine = None
        self.samples = {}
        self.match_threshold = match_treshold
        self.match_filter = match_filter
        self.token_size = token_size

    def set_engine(self, engine):
        self.engine = engine

    def set_kb(self, kb):
        kb.query_samples = self.samples

    def load_kb(self, kb):
        self.samples = kb.query_samples

    def set_child_heuristics(self, heuristics):
        self.child_heuristics = heuristics

    async def after_response(self, entry):
        url = urlparse(entry.request.url)
        if len(url.query) > 0:
            sample_simhash = await self._get_sample(url)
            if self._match(entry.response, sample_simhash):
                raise RejectRequest("Junk query response")

    async def _get_sample(self, parsed_url):
        sample_key = parsed_url.netloc + parsed_url.path
        try:
            return self.samples[sample_key]
        except KeyError:
            random_query = "{scheme}://{netloc}{path}?{query}"\
                .format(scheme=parsed_url.scheme, netloc=parsed_url.netloc, path=parsed_url.path, query=str(uuid4()))
            sample = await self.engine.perform_high_priority(Entry.create(random_query), self.child_heuristics)
            self.samples[sample_key] = self._hash_response(sample.response)
            return self.samples[sample_key]

    def _match(self, response, sample_simhash):
        if "md5" in sample_simhash:
            if self._is_text(response.raw):
                return False
            else:
                return hashlib.md5(response.raw).digest() == sample_simhash["md5"]
        elif "simhash" in sample_simhash:
            try:
                response_simhash = self._create_simhash(response.content)
                return self._simhash_equal(response_simhash, self._create_simhash(sample_simhash["simhash"]))
            except UnicodeDecodeError:
                return False

    def _hash_response(self, response):
        try:
            return {"simhash": self._create_simhash(response.content).value}
        except UnicodeDecodeError:
            return {"md5": hashlib.md5(response.raw).digest()}

    def _is_text(self, raw_response):
        try:
            raw_response.decode("utf8")
            return True
        except UnicodeDecodeError:
            return False

    def _create_simhash(self, content_response):
        return Simhash(content_response, self.match_filter, self.token_size)

    def _simhash_equal(self, simhash, other_simhash):
        return simhash.distance(other_simhash) <= self.match_threshold



# Tachyon - Fast Multi-Threaded Web Discovery Tool
# Copyright (c) 2011 Gabriel Tremblay - initnull hat gmail.com
# Copyright (C) 2018-  Delve Labs inc.
#
# GNU General Public Licence (GPL)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA


from urllib.parse import urljoin

from hammertime.rules.deadhostdetection import OfflineHostException
from hammertime.ruleset import StopRequest, RejectRequest

from .textutils import output_manager, PrettyOutput
from .result import ResultAccumulator


class FileFetcher:

    def __init__(self, host, hammertime, accumulator=None):
        self.host = host
        self.hammertime = hammertime
        self.accumulator = accumulator or ResultAccumulator(output_manager=output_manager or PrettyOutput())

    async def fetch_files(self, file_list):
        for file in file_list:
            url = urljoin(self.host, file["url"])
            self.hammertime.request(url, arguments={"file": file})

        iterator = self.hammertime.successful_requests()

        # Really make sure we are done (issue in hammertime 0.5.1 when first request is a failure?)
        while iterator.has_pending():
            async for entry in iterator:
                try:
                    self.accumulator.add_entry(entry)
                except OfflineHostException:
                    raise
                except RejectRequest:
                    pass
                except StopRequest:
                    continue


class ValidateEntry:
    """
    Combines the RejectSoft404 and RejectErrorBehavior, but excludes problems when
    the result requested a string match and found it.
    """

    async def after_response(self, entry):
        if entry.result.string_match:
            # We found what we were looking for, this entry has to be valid.
            return

        if getattr(entry.result, "error_behavior", False):
            raise StopRequest("Error behavior detected.")
        if getattr(entry.result, "soft404", False):
            raise RejectRequest("Soft 404 detected")

    async def on_request_successful(self, entry):
        # Apply the same logic as on response
        await self.after_response(entry)
